/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you add "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * 24h timer with web interface.
 * A button connected to PD6 may be used to manually switch the relay.
 * Two DC18S20 sensors connected with a pull-up to PD4 to monitor temperature.
 * 
 * See http://tuxgraphics.org/electronics/
 *
 * Chip type: Atmega328  or atmega644 with ENC28J60
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "net.h"
#include "websrv_help_functions.h"
#include "onewire.h"
#include "ds18x20.h"

// set output to VCC, red LED off
#define LEDOFF PORTB|=(1<<PORTB1)
// set output to GND, red LED on
#define LEDON PORTB&=~(1<<PORTB1)
// to test the state of the LED
#define LEDISOFF PORTB&(1<<PORTB1)
//
#define RELAYON PORTD|=(1<<PORTD7)
#define RELAYOFF PORTD&=~(1<<PORTD7)
#define RELAYISON PORTD&(1<<PORTD7)
#define RELAY_D_ON relaystate=1;relay_switch_delay=0x2fff
#define RELAY_D_OFF relaystate=0;relay_switch_delay=0x2fff

// please modify the following lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices. The IP address may be changed at run-time but the
// MAC address is hardcoded:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
//
// The IP of this device (can also be changed at run-time, see README file):
static uint8_t myip[4] = {10,0,0,29};
//static uint8_t myip[4] = {192,168,0,81};
// listen port for tcp/www:
static uint16_t mywwwport=80;
//static uint16_t mywwwport=81;
// the password string (only characters: a-z,0-9,_,.,-,# ), max len 8 char:
static char password[9]="secret";
static uint8_t pool_sensor_number=0; // can be 0 or 1, the sensor which is recorded 5min after pump on.
// set this to one if you want the graphs in farenheit
static uint8_t graph_in_f=0;
// how often (in minutes times 10, 12=120min, 24=240min, 36=6h 72=12h) to record the data to a graph (max value=255 ):
static uint8_t rec_interval=36;
//
// You can calibrate the clock here a bit.
// The number can be positive or negative and says
// how many seconds to add or subtract every 240h hours.
// I recommend to let the board run for a month before you
// touch this number.
#define CLOCK_CAL 1
// To make it easier to calibrate the system you can have the system 
// show seconds on the main page. Note that seconds jump once in a while.
//#define CLOCK_SHOW_SEC 
// -------------- do not change anything below this line ----------
static uint8_t pool_sensor_time[3]={0,0,0}; // when pool sensor was last read, hour=0, min=1, sec=2
static uint16_t pool_sensor_hours_last_read=0xffff;
static int16_t pool_sensor_Tempdata; // temperature times 10
static char label[2][8]={"air","air"}; // we write pool into one of them during initialisation
//
#define MAXSENSORS 3
static uint8_t gSensorIDs[MAXSENSORS][OW_ROMCODE_SIZE];
static int16_t gTempdata[MAXSENSORS]; // temperature times 10
static int8_t gNsensors=0;
static volatile uint8_t gMeasurementTimer=0; // sec
static uint8_t gTemp_measurementstatus=0; // 0=ok,1=error
// The buffer is the packet size we can handle and its upper limit is
// given by the amount of RAM that the atmega chip has. 
#define BUFFER_SIZE 995
static uint8_t buf[BUFFER_SIZE+1];

// global string buffer
#define STR_BUFFER_SIZE 40
static char gStrbuf[STR_BUFFER_SIZE+1];
//
static uint16_t gPlen;
static uint16_t tablebitmap[6]={0,0,0,0,0,0}; // each of the 96 bits represents one field in the time table
static volatile uint8_t time[3]={0,0,0}; // hour=0, min=1, sec=2
static uint8_t relaystate=0; // 1 = relay is on, 0 = off
static uint16_t relay_switch_delay=0;
//
#define HIST_BUFFER_SIZE 50
// position at which the EEPROM is used for storage of temp. data:
#define EEPROM_TEMP_POS_OFFSET 40 
static uint8_t gTemphistnextptr=0;
static uint16_t gRecMin=0; // minutes counter for recording of history graphs, set to 0 when graph stored and then count up every minute. 
// 
#define USERAMDAT 1
#ifdef USERAMDAT
// 2 sensors max:
static uint16_t ram_dat[2][HIST_BUFFER_SIZE];
int16_t read_dat(uint8_t pos,uint8_t sensor){
        return(ram_dat[sensor][pos]);
}
void store_dat(int16_t val, uint8_t pos,uint8_t sensor)
{
        ram_dat[sensor][pos]=val;
}
#else //USERAMDAT
// 
// sensor=0..2
int16_t read_dat(uint8_t pos,uint8_t sensor)
{
        return ((int16_t)eeprom_read_word((uint16_t *)(2*(uint16_t)pos+EEPROM_TEMP_POS_OFFSET+(2*HIST_BUFFER_SIZE*(uint16_t)sensor))));
}

// no overflow protection here make sure you stay within the available eeprom size:
void store_dat(int16_t val, uint8_t pos,uint8_t sensor)
{
        eeprom_write_word((uint16_t *)(2*(uint16_t)pos+EEPROM_TEMP_POS_OFFSET+(2*HIST_BUFFER_SIZE*(uint16_t)sensor)),(uint16_t)val); 
}
#endif //USERAMDAT

void init_store_dat(void)
{
        uint8_t i=0;
        while(i<HIST_BUFFER_SIZE){
                store_dat(0,i,0);
                store_dat(0,i,1);
                i++;
        }
}
// called when we want to store a sample 
void store_graph_dat(void){
        int8_t i,j;
        int16_t tdat;
        gRecMin=0;
        j=gNsensors;
        if (j>3){
                j=3;
        }
        for ( i=0; i<j; i++ ) {
                tdat=gTempdata[i];
                if (i==pool_sensor_number){
                        tdat=pool_sensor_Tempdata;
                }
                store_dat(tdat,gTemphistnextptr,i);
        }
        gTemphistnextptr++;
        gTemphistnextptr=gTemphistnextptr%HIST_BUFFER_SIZE;
}

//
uint8_t verify_password(char *str)
{
        // the first characters of the received string are
        // a simple password/cookie:
        if (strncmp(password,str,strlen(password))==0){
                return(1);
        }
        return(0);
}

uint16_t http200ok(uint8_t nocache)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n"));
        if (nocache){
                plen=fill_tcp_data_p(buf,plen,PSTR("Pragma: no-cache\r\n"));
        }
        return(fill_tcp_data_p(buf,plen,PSTR("\r\n")));
}


uint16_t http200okjs(uint8_t nocache)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: application/x-javascript\r\n"));
        if (nocache){
                plen=fill_tcp_data_p(buf,plen,PSTR("Pragma: no-cache\r\n"));
        }
        return(fill_tcp_data_p(buf,plen,PSTR("\r\n")));
}


uint16_t fill_tcp_data_uint8_pad(uint8_t *buf,uint16_t plen,uint8_t i)
{
        itoa(i,gStrbuf,10); // convert integer to string
        if (i<10) plen=fill_tcp_data(buf,plen,"0");
        return(fill_tcp_data(buf,plen,gStrbuf));
}

uint16_t fill_tcp_data_int(uint8_t *buf,uint16_t plen,uint16_t ui)
{
        uint8_t i=6;
        if (ui==0){
                return(fill_tcp_data(buf,plen,"0"));
        }
        // convert a 16bit unsigned integer into a printable string, avr-lib has
        // not function for that therefore we do it step by step:
        gStrbuf[i]='\0';
        while(ui&&i){
                i--;
                gStrbuf[i]=(char)((ui%10 & 0xf))+0x30;
                ui=ui/10;
        }
        return(fill_tcp_data(buf,plen,&gStrbuf[i]));
}

// writes to global array gSensorIDs
int8_t search_sensors(void)
{
	uint8_t diff;
        uint8_t nSensors=0;
	for( diff = OW_SEARCH_FIRST; 
		diff != OW_LAST_DEVICE && nSensors < MAXSENSORS ; )
	{
		DS18X20_find_sensor( &diff, &(gSensorIDs[nSensors][0]) );
		
		if( diff == OW_PRESENCE_ERR ) {
			return(-1); //No Sensor found
		}
		if( diff == OW_DATA_ERR ) {
			return(-2); //Bus Error
		}
		nSensors++;
	}
	return nSensors;
}

// start a measurement for all sensors on the bus:
void start_temp_meas(void){
        gTemp_measurementstatus=0;
        if ( DS18X20_start_meas(NULL) != DS18X20_OK) {
                gTemp_measurementstatus=1;
        }
}

// insert a '.' before the last digit/character.
// "100" will become "10.0"
// This function writes to gStrbuf
// The user is responsible for allocating enough space.
void adddecimalpoint2(int16_t num){
        uint8_t i;
        itoa(num/10,gStrbuf,10); 
        i=strlen(gStrbuf);
        gStrbuf[i]='.';
        itoa(abs(num)%10,&(gStrbuf[i+1]),10); 
}

// convert celsius times 10 values to Farenheit times 10
int16_t c2f10(int16_t celsiustimes10){
        return(((celsiustimes10 * 18)/10) + 320);
}


// read the latest measurement off the scratchpad of the ds18x20 sensor
// and store it in gTempdata
void read_temp_meas(void){
        uint8_t i;
        uint8_t subzero, cel, cel_frac_bits;
        for ( i=0; i<gNsensors; i++ ) {
        
                if ( DS18X20_read_meas( &gSensorIDs[i][0], &subzero,
                                &cel, &cel_frac_bits) == DS18X20_OK ) {
                        gTempdata[i]=cel*10;
                        gTempdata[i]+=DS18X20_frac_bits_decimal(cel_frac_bits);
                        if (subzero){
                                gTempdata[i]=-gTempdata[i];
                        }
                }else{
                        gTempdata[i]=0;
                }
        }
}

/***************
viewport tag properties (see as well
http://developer.apple.com/library/safari/#documentation/appleapplications/refer
+ence/SafariHTMLRef/Articles/InputTypes.html#//apple_ref/doc/uid/TP40008055-SW3
)

width - The width of the viewport in pixels. The default is 980. The
range is from 200 to 10,000.
You can as well set it to the special value "device-width"

height - The height of the viewport in pixels. The default is calculated
based on the value of the width property and the aspect ratio of the
device. The range is from 223 to 10,000 pixels.
You can as well set it to the special value "device-height"

initial-scale - The initial scale of the viewport as a multiplier. The
default is calculated to fit the webpage in the visible area. The range
is determined by the minimum-scale and maximum-scale properties.

You can set only the initial scale of the viewport?the scale of the
viewport the first time the webpage is displayed. Thereafter, the user
can zoom in and out unless you set user-scalable to no. Zooming by the
user is also limited by the minimum-scale and maximum-scale properties.

minimum-scale - Specifies the minimum scale value of the viewport. The
default is 0.25. The range is from >0 to 10.0.

maximum-scale - Specifies the maximum scale value of the viewport. The
default is 1.6. The range is from >0 to 10.0.

user-scalable - Determines whether or not the user can zoom in and
out?whether or not the user can change the scale of the viewport. Set to
yes to allow scaling and no to disallow scaling. The default is yes.

Setting user-scalable to no also prevents a webpage from scrolling when
entering text in an input field.

****************/
// there is unfortunately one problem: Differnet mobile browsers
// interprete the same syntax in different ways. Especially Opera
// is quite off from safari and the android browser. The only
// thing that works well is width=device-width
uint16_t print_line_mobile_webpage(uint16_t plen)
{
        plen=fill_tcp_data_p(buf,plen,PSTR("<meta name=viewport content=\"width=device-width\">\n"));
        return(plen);
}

uint16_t print_webpage_setclock(void)
{
        uint16_t plen;
        plen=http200ok(0);
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Set clock</h2>\n<pre>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<form  action=cc method=get>\n\
time :<input type=text name=tv id=tf size=10>\n\
passw:<input type=password name=pw size=10><input type=submit value=\"set time\">\n\
</form>\n\
<script>\n\
d=new Date();\n\
e=document.getElementById(\"tf\");\n\
m=d.getMinutes();\n\
s= (m<10) ? \"0\" : \"\";\n\
e.value=d.getHours()+\":\"+s+d.getMinutes();\n\
</script>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("</pre><hr>\n"));
        return(plen);
}

// p3.js
uint16_t print_p3_js(void)
{
        // I am sorry that the javascript is a bit unreadable 
        // but we absolutely need to keep the amount of code
        // in bytes small
        uint16_t plen;
        plen=http200okjs(0);
        // g = graphdata array
        // m = minutes since last data, global variable
        // ri = recording interval in min, global variable
        // d = new Date(), global variable
        plen=fill_tcp_data_p(buf,plen,PSTR("\
function pad(n){\
var s=n.toString(10);\
if (s.length<2){return(\"0\"+s)}\
return(s);\
}\n\
function dw(s){document.writeln(s)}\n\
function scsv(g1,g2){\
document.open();\
dw(\"<pre>#Copy/paste as txt file:\");\
"));
        plen=fill_tcp_data_p(buf,plen,PSTR("dw(\"#date;"));
        plen=fill_tcp_data(buf,plen,label[0]);
        plen=fill_tcp_data(buf,plen,";");
        plen=fill_tcp_data(buf,plen,label[1]);
        plen=fill_tcp_data(buf,plen,";");
        plen=fill_tcp_data_p(buf,plen,PSTR("\");\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
var i,ms,pt;\
ms=d.getTime();\
var l=new Date();\
for(i=0;i<g1.length;i++){\
pt=ms-((m+ri*i)*60*1000);\
l.setTime(pt);\
dw(l.getFullYear()+\"-\"+pad(l.getMonth()+1)+\"-\"+pad(l.getDate())+\" \"+pad(l.getHours())+\":\"+pad(l.getMinutes())+\";\"+g1[i]/10+\";\"+g2[i]/10+\";\");\
}\
document.close();\
}\n\
"));
        return(plen);
}

// p2.js
uint16_t print_p2_js(void)
{
        // I am sorry that the javascript is a bit unreadable 
        // but we absolutely need to keep the amount of code
        // in bytes small
        uint16_t plen;
        plen=http200okjs(0);
        // gd = graphdata array
        // c = color
        // a = amplification factor
        // h = heading
        // o = offset
        //
        // m = minutes since last data, global variable
        // ri = recording interval in min, global variable
        // d = new Date(), global variable
        //
        // pbar(w,c,ts) w=width c=color ts=time
        plen=fill_tcp_data_p(buf,plen,PSTR("\
function pbar(w,c,ts){\
if(w<5){w=5;}\
dw(\"<p style=\\\"width:\"+w+\"px;background-color:#\"+c+\";margin:2px;border:1px #999 solid;white-space:pre;\\\">\"+ts+\"</p>\");\
}\n\
function bpt(gd,c,a,o,h){\
dw(\"<h2>\"+h+\"</h2>\");\
var i,n,ms,pt;\
ms=d.getTime();\
var l=new Date();\
for(i=0;i<gd.length;i++){\
 pt=ms-((m+ri*i)*60*1000);\
 l.setTime(pt);\
 n=gd[i];\
 pbar(parseInt(a*n+o),c,pad(l.getDate())+\"-\"+pad(l.getHours())+\":\"+pad(l.getMinutes())+\"=\"+(parseFloat(n)/10).toFixed(1));\
}\n\
dw(\"<br>Format: day-hh:mm=val<br>\");\
}\n\
"));
        return(plen);
}

// sensor = 0 or 1
uint16_t print_webpage_graph_array(uint8_t *buf,uint8_t sensor)
{
        int8_t i;
        uint8_t nptr;
        uint16_t plen;
        plen=http200okjs(1);
        nptr=gTemphistnextptr; // gTemphistnextptr may change so we make a snapshot here
        i=nptr;
        plen=fill_tcp_data_p(buf,plen,PSTR("gdt"));
        plen=fill_tcp_data_int(buf,plen,sensor);
        plen=fill_tcp_data_p(buf,plen,PSTR("=Array("));
        if (!eeprom_is_ready()){
                // don't sit and wait 
                plen=fill_tcp_data_p(buf,plen,PSTR("0);\n"));
                return(plen);
        }
        while(1){
                if (i!=nptr){ // not first time
                        plen=fill_tcp_data_p(buf,plen,PSTR(","));
                }
                i--;
                if (i<0){
                        i=HIST_BUFFER_SIZE-1;
                }
                if (graph_in_f){
                        plen=fill_tcp_data_int(buf,plen,c2f10(read_dat(i,sensor)));
                }else{
                        plen=fill_tcp_data_int(buf,plen,read_dat(i,sensor));
                }
                if (i==nptr){
                        break; // end loop
                }
        }
        plen=fill_tcp_data_p(buf,plen,PSTR(");\n"));
        return(plen);
}

uint16_t print_webpage_graph(void)
{
        uint16_t plen;
        uint8_t th,tm;
        plen=http200ok(1);
        th=time[0];tm=time[1]; // copy immediately
        // global javascript variables:
        plen=fill_tcp_data_p(buf,plen,PSTR("<script>var d=new Date();var m="));
        plen=fill_tcp_data_int(buf,plen,gRecMin);
        plen=fill_tcp_data_p(buf,plen,PSTR(";var ri="));
        plen=fill_tcp_data_int(buf,plen,rec_interval);
        plen=fill_tcp_data_p(buf,plen,PSTR("0;\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("d.setHours("));
        itoa(th,gStrbuf,10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR(");\nd.setMinutes("));
        itoa(tm,gStrbuf,10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data(buf,plen,");\n");
        plen=fill_tcp_data_p(buf,plen,PSTR("</script>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<script src=/p3.js></script><script src=/p2.js></script>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<script src=/gdt0.js></script>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<script src=/gdt1.js></script>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/td>[back]</a> "));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=javascript:scsv(gdt0,gdt1)>[CSV data]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<script>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("dw(\"<table border=1><tr><td>\");\n"));
        if (graph_in_f){
                plen=fill_tcp_data_p(buf,plen,PSTR("bpt(gdt0,\"ac2\",0.3,25,\""));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("bpt(gdt0,\"ac2\",0.7,130,\""));
        }
        plen=fill_tcp_data(buf,plen,label[0]);
        plen=fill_tcp_data_p(buf,plen,PSTR("\");\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("dw(\"</td><td>\");\n"));
        // orange: eb0 green: ac2 blue: 78c orange2: f91 red: f00
        if (graph_in_f){
                plen=fill_tcp_data_p(buf,plen,PSTR("bpt(gdt1,\"f91\",0.3,25,\""));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("bpt(gdt1,\"f91\",0.7,130,\""));
        }
        plen=fill_tcp_data(buf,plen,label[1]);
        plen=fill_tcp_data_p(buf,plen,PSTR("\");\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("dw(\"</td></tr></table>\");\n</script>\n<br><hr>\n"));
        return(plen);
}

uint16_t print_webpage_sensor(void)
{
        uint16_t plen;
        int8_t i=0;
        plen=http200ok(1);
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a> <a href=/gp>[graph]</a> <a href=/td>[refresh]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Temperature</h2>\n<pre>\n"));
        // max 3 sensors otherwise this page might be bigger than the buf variable:
        if (gNsensors<1){
                plen=fill_tcp_data_p(buf,plen,PSTR("no sensor found"));
        }
        while(i<gNsensors && i<3){ 
                if (i<2){
                        plen=fill_tcp_data(buf,plen,label[i]);
                }else{
                        plen=fill_tcp_data_p(buf,plen,PSTR("temp"));
                        plen=fill_tcp_data_int(buf,plen,i);
                }
                plen=fill_tcp_data_p(buf,plen,PSTR(": "));
                if (i==pool_sensor_number){
                        if (pool_sensor_hours_last_read==0xffff){
                                plen=fill_tcp_data_p(buf,plen,PSTR("(waiting for waterflow/pump"));
                        }else{

                                adddecimalpoint2(pool_sensor_Tempdata); plen=fill_tcp_data(buf,plen,gStrbuf);
                                plen=fill_tcp_data_p(buf,plen,PSTR("&deg;C ["));
                                adddecimalpoint2(c2f10(pool_sensor_Tempdata)); plen=fill_tcp_data(buf,plen,gStrbuf);
                                plen=fill_tcp_data_p(buf,plen,PSTR("&deg;F] ("));
                                if (pool_sensor_hours_last_read<24){
                                        plen=fill_tcp_data_p(buf,plen,PSTR("time "));
                                        plen=fill_tcp_data_uint8_pad(buf,plen,pool_sensor_time[0]);
                                        plen=fill_tcp_data(buf,plen,":");
                                        plen=fill_tcp_data_uint8_pad(buf,plen,pool_sensor_time[1]);
                                }else{
                                        plen=fill_tcp_data_int(buf,plen,pool_sensor_hours_last_read);
                                        plen=fill_tcp_data_p(buf,plen,PSTR("h ago"));
                                }
                        }
                        plen=fill_tcp_data_p(buf,plen,PSTR(")\n pipe:"));

                }
                adddecimalpoint2(gTempdata[i]); plen=fill_tcp_data(buf,plen,gStrbuf);
                plen=fill_tcp_data_p(buf,plen,PSTR("&deg;C ["));
                adddecimalpoint2(c2f10(gTempdata[i])); plen=fill_tcp_data(buf,plen,gStrbuf);
                plen=fill_tcp_data_p(buf,plen,PSTR("&deg;F] (now)\n"));
                i++;
        }
        if (gTemp_measurementstatus==1){
                plen=fill_tcp_data_p(buf,plen,PSTR("warning: sensor error\n"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("</pre>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<small>pool temp. = pipe temp. when pump on for 5min or more.</small><br>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<hr>\n"));
        return(plen);
}

uint16_t print_webpage_relay(void)
{
        uint16_t plen;
        plen=http200ok(1);
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a> <a href=/sw>[refresh]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Pump Switch</h2>\n<pre>state: "));
        if (relaystate==1){
                plen=fill_tcp_data_p(buf,plen,PSTR("<font color=#009900>ON</font>"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("OFF"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\n<form action=/cr method=get>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<input type=hidden name=sw value="));
        if (relaystate==1){
                plen=fill_tcp_data_p(buf,plen,PSTR("0>\npassw: <input type=password size=10 name=pw>"));
                plen=fill_tcp_data_p(buf,plen,PSTR("<input type=submit value=\"switch off\"></form>\n"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("1>\npassw: <input type=password size=10 name=pw>"));
                plen=fill_tcp_data_p(buf,plen,PSTR("<input type=submit value=\"switch on\"></form>\n"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("</pre><hr>\n"));
        return(plen);
}

// check the tablebitmap and return:
// 0 there is no change compared to previous minute and current state is off
// 1 the on/off state changed from on to off
// 2 the on/off state changed from off to on
// 3 there is no change compared to previous minute and current state is on
uint8_t chk_tablebitmap(uint8_t h, uint8_t m)
{
        uint8_t tablepage,bitpos;
        uint8_t curr=0;
        uint8_t prev=0;
        tablepage=h/4;// which table page is the current
        if (tablepage>5) tablepage=5; // bounds check in case of garbage time
        bitpos=((h%4)*60+m)/15;
        if (bitpos>15) bitpos=15; // bounds check in case of garbage time
        if (tablebitmap[tablepage] & (1<<bitpos)){
                curr=1;
        }
        // now check what the previous value was
        if (m>0){
                m--;
        }else{
                m=59;
                if (h>0){
                        h--;
                }else{
                        h=23;
                }
        }
        tablepage=h/4;// which table page is the current
        if (tablepage>5) tablepage=5; // bounds check in case of garbage time
        bitpos=((h%4)*60+m)/15;
        if (bitpos>15) bitpos=15; // bounds check in case of garbage time
        if (tablebitmap[tablepage] & (1<<bitpos)){
                prev=1;
        }
        if (prev==curr && curr==0) return(0);
        if (prev==curr && curr==1) return(3);
        if (curr==1 && prev==0) return(2);
        return(1);
}

void relay_change_check(void){
        uint8_t i;
        i=chk_tablebitmap(time[0],time[1]);
        if (i==2){
                RELAY_D_ON;
        }
        if (i==1){
                RELAY_D_OFF;
        }
}

void relay_on_off_check(void){
        uint8_t i;
        i=chk_tablebitmap(time[0],time[1]);
        if (i==2||i==3){
                RELAY_D_ON;
        }else{
                RELAY_D_OFF;
        }
}

// p1.js
uint16_t print_p1_js(void)
{
        // I am sorry that the javascript is a bit unreadable 
        // but we absolutely need to keep the amount of code
        // in bytes small.
        //
        // In the function fill we encode all the 96 fiels into 6 numbers 
        // with 16 bit. Every bit represents a table box. If the box was 
        // green then the bit is 1 otherwise zero.
        uint16_t plen;
        plen=http200okjs(0);
        plen=fill_tcp_data_p(buf,plen,PSTR("\
var col=\"\";\n\
var cc=0;\n\
var e,i,j;\n\
function dw(s){document.writeln(s)}\n\
function wr(t,i){dw(\"<tr><td>\"+t+\"</td><td id=t\"+i+\" onmousedown=cbg(this) onmouseover=mc(this)>&nbsp;</td></tr>\")}\n\
function dwh(s){dw(\"<table border=1><tr><td>Time</td><td>green=on<br>white=off</td></tr>\")}\n\
function cbg(e) {\n\
cc=1;\n\
col=e.style.backgroundColor ? \"\" : \"#7AFA7A\";\n\
e.style.backgroundColor=col;\n\
}\n\
function sc(){cc=0}\n\
function fill(ff)\n\
{\n\
var b=new Array();\n\
for(i=0;i<6;i++){\n\
b[i]=0;\n\
for(j=0;j<16;j++){\n\
e=document.getElementById(\"t\"+(i*16+j));\n\
if (e.style.backgroundColor) b[i]|=1<<j;\n\
}\n\
}\n\
ff.elements[0].value=b[0] +\"-\"+b[1]+\"-\"+b[2]+\"-\"+b[3]+\"-\"+b[4]+\"-\"+b[5]+\"-\";\n\
}\n\
function it(a)\n\
{\n\
for(i=0;i<6;i++){\n\
for(j=0;j<16;j++){\n\
e=document.getElementById(\"t\"+(i*16+j));\n\
if (a[i]&(1<<j)) e.style.backgroundColor=\"#7AFA7A\";\n\
}\n\
}\n\
}"));
        return(plen);
}


// this the top level main web page
uint16_t print_webpage_main(void)
{
        uint16_t plen;
        int8_t i;
        plen=http200ok(1);
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/sc>[set]</a> <a href=/td>[temp]</a> <a href=/sw>[switch]</a> <a href=/>[refresh]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h1>Pool Timer</h1>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("Current time:"));
        plen=fill_tcp_data_uint8_pad(buf,plen,time[0]);
        plen=fill_tcp_data(buf,plen,":");
        plen=fill_tcp_data_uint8_pad(buf,plen,time[1]);
#ifdef CLOCK_SHOW_SEC
        plen=fill_tcp_data(buf,plen,":");
        plen=fill_tcp_data_uint8_pad(buf,plen,time[2]);
#endif
        plen=fill_tcp_data_p(buf,plen,PSTR(" Pump: "));
        if (relaystate==1){
                plen=fill_tcp_data_p(buf,plen,PSTR("<font color=#009900>ON</font>"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("off"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<br>Click fields to change.\n\
<script src=p1.js></script>\n\
<script>\n\
function mc(e){if (cc) e.style.backgroundColor=col}\n\
dw(\"<table onmouseup=sc()>\");\n\
dw(\"<tr><td>\");\n\
dwh();\n\
var i,t;\n\
for(i=0;i<48;i++){\n\
t=\"\";\n\
if (i%4==0) t=i/4+\":00\";\n\
wr(t,i);\n\
}\n\
dw(\"</table></td><td>\");\n\
dwh();\n\
for(i=48;i<96;i++){\n\
t=\"\";\n\
if (i%4==0) t=i/4+\":00\";\n\
wr(t,i);\n\
}\n\
dw(\"</table></td></tr></table>\");\n\
dw(\"1 row=15min\");\n\
"));
        // initialize the table with the current settings
        plen=fill_tcp_data_p(buf,plen,PSTR("it(new Array("));
        plen=fill_tcp_data_int(buf,plen,tablebitmap[0]);
        i=1;
        while(i<6){
                plen=fill_tcp_data(buf,plen,",");
                plen=fill_tcp_data_int(buf,plen,tablebitmap[i]);
                i++;
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("));\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
</script>\n\
<br><br>\n\
<form onsubmit=fill(this) action=st method=get>\n\
<input type=hidden name=tt>\n\
passw: <input type=password size=10 name=pw>\n\
<input type=submit value=save>\n\
</form>\n\
<hr>(c)tuxgraphics\n\
"));
        // debug dbg
        //plen=fill_tcp_data_int(buf,plen,chk_tablebitmap(time[0],time[1]));
        return(plen);
}

// takes a string of the form command/Number and analyse it (e.g "cc?tv=8%3A15&pw=secret HTTP/1.1")
// The first char of the url ('/') is already removed.
int8_t analyse_get_url(char *str)
{
        int8_t i;
        int8_t j;
        uint8_t n;
        if (str[0]==' '){ // end of url, main page
                return(1); 
        }
        // --------
        if (strncmp("favicon.ico",str,11)==0){
                gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 301 Moved Permanently\r\nLocation: "));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("http://tuxgraphics.org/ico/clk.ico"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("\r\n\r\nContent-Type: text/html\r\n\r\n"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>301 Moved Permanently</h1>\n"));
                return(10);
        }
        // --------
        if (strncmp("sc",str,2)==0){
                gPlen=print_webpage_setclock();
                return(10);
        }
        if (strncmp("td",str,2)==0){
                gPlen=print_webpage_sensor();
                return(10);
        }
        if (strncmp("sw",str,2)==0){
                gPlen=print_webpage_relay();
                return(10);
        }
        if (strncmp("p1.js",str,5)==0){
                gPlen=print_p1_js();
                return(10);
        }
        if (strncmp("p2.js",str,5)==0){
                gPlen=print_p2_js();
                return(10);
        }
        if (strncmp("p3.js",str,5)==0){
                gPlen=print_p3_js();
                return(10);
        }
        if (strncmp("gp",str,2)==0){
                gPlen=print_webpage_graph();
                return(10);
        }
        if (strncmp("gdt0.js",str,7)==0){
                gPlen=print_webpage_graph_array(buf,0);
                return(10);
        }
        if (strncmp("gdt1.js",str,7)==0){
                gPlen=print_webpage_graph_array(buf,1);
                return(10);
        }
        // --------
        if (strncmp("st",str,2)==0){ // save the tablebitmap
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(gStrbuf);
                        if (verify_password(gStrbuf)){
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"tt")){
                                        urldecode(gStrbuf);
                                        gStrbuf[37]='\0'; // max length
                                        // gStrbuf is a string like this: 11-58-0-0-0-0-
                                        // no space allowed
                                        if (!isdigit(gStrbuf[0])){
                                                return(0);
                                        }
                                        n=0;i=0;j=0;
                                        while(gStrbuf[i] && n<6){
                                                if(gStrbuf[i]=='-'){
                                                        gStrbuf[i]='\0';
                                                        tablebitmap[n]=atol(&gStrbuf[j]);
                                                        n++;
                                                        j=i+1;
                                                }
                                                i++;
                                        }
                                        relay_on_off_check();
                                        return(1); // main page
                                }
                        }
                }
                return(-1);
        }
        if (strncmp("cc",str,2)==0){ // change clock
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(gStrbuf);
                        if (verify_password(gStrbuf)){
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"tv")){
                                        urldecode(gStrbuf);
                                        gStrbuf[5]='\0';
                                        // gStrbuf is a string like this: 11:58
                                        // no space allowed
                                        if (!isdigit(gStrbuf[0])){
                                                return(0);
                                        }
                                        // find position of ':' in string:
                                        i=0;
                                        if (gStrbuf[1]==':'){
                                                i=1;
                                        }
                                        if (gStrbuf[2]==':'){
                                                i=2;
                                        }
                                        if (i){
                                                gStrbuf[i]='\0';
                                                time[0]=atoi(gStrbuf);
                                                time[1]=atoi(&gStrbuf[i+1]);
                                                time[2]=0; 
                                                chk_tablebitmap(time[0],time[1]);
                                                relay_on_off_check();
                                                return(1); // main page
                                        }
                                        return(0);
                                }
                        }
                }
                return(-1);
        }
        i=0; // switch on or off 
        if (strncmp("cr",str,2)==0){ // change relay state FORM
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"sw")){
                        if (gStrbuf[0]=='1'){
                                i=1;
                        }
                        if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                                urldecode(gStrbuf);
                                if (verify_password(gStrbuf)){
                                        if (i){
                                                RELAY_D_ON;
                                        }else{
                                                RELAY_D_OFF;
                                        }
                                        return(1); // main page
                                }
                        }
                        return(-1);
                }
                return(0);
        }
        return(0);
}

// called when TCNT2==OCR2A
// that is in 50Hz intervals (49.8246Hz intervals, clock is a bit behind)
// This is used as a "clock" to store the data
ISR(TIMER2_COMPA_vect){
        static uint8_t cnt2step=0;
        cnt2step++;
        if (cnt2step>49){ 
                gMeasurementTimer++;
                time[2]++; // step sec
                cnt2step=0;
        }
}

/* setup timer T2 as an interrupt generating time base.
* You must call once sei() in the main program */
void init_cnt2(void)
{
        PRR&=~(1<<PRTIM2); // write power reduction register to zero
        TIMSK2=(1<<OCIE2A); // compare match on OCR2A
        TCNT2=0;  // init counter
        OCR2A=244; // value to compare against (we divide by 245)
        TCCR2A=(1<<WGM21); // do not change any output pin, clear at compare match
        // divide clock by 1024: 12.5MHz/1024=12207.0313 Hz
        TCCR2B=(1<<CS22)|(1<<CS21)|(1<<CS20); // clock divider, start counter
        // 12207.0313 / 245= 49.8246Hz
}


int main(void){
        uint16_t dat_p;
        int8_t i=0;
        uint8_t days=0;
        uint16_t debounce=0;
        uint8_t gratutious_arp_cnt=0;
        int8_t relay_on_5_min=0;
        static int8_t state=0;
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(0); // 60us

        // enable PD6, as input, change state of relay manually
        DDRD&= ~(1<<DDD6);
        PORTD|= (1<<PIND6); // internal pullup resistor on, jumper goes to ground
        //initialize enc28j60
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(0); // 60us
        gRecMin=((uint16_t)rec_interval*10)-1; // store first sample one minute after power on
        // initialize graph:
        init_store_dat();
        // time keeping
        init_cnt2();
        sei();
        // LED
        DDRB|= (1<<DDB1); // enable PB1, LED as output
        LEDOFF;

        // the transistor on PD7 (relay)
        DDRD|= (1<<DDD7);
        RELAYOFF;
        // Magjack leds configuration, see enc28j60 datasheet, page 11 
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        // label the pool sensor
        if (pool_sensor_number<2){
                strcpy(label[pool_sensor_number],"pool");
        }

        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,mywwwport);
        // ds18s20 sensors:
        gNsensors=search_sensors();
        if (gNsensors>0){
                start_temp_meas();
        }
        gMeasurementTimer=0; // set again to zero before starting while loop
        while(1){
                // get the next new packet:
                gPlen = enc28j60PacketReceive(BUFFER_SIZE, buf);
                dat_p=packetloop_icmp_tcp(buf,gPlen);

                //dat_p will ne unequal to zero if there is a valid packet (without crc error) 
                if(dat_p==0){
                        // no packets, do time keeping and other tasks
                        //
                        // check relay state and set relay accordingly
                        // we do this here to avoid sending a packet at 
                        // the same time as switching the relay:
                        if (relay_switch_delay==1){
                                relay_switch_delay=0;
                                if (RELAYISON){
                                        if (relaystate==0) {RELAYOFF;LEDOFF;}
                                }else{
                                        if (relaystate==1) {RELAYON;LEDON;}
                                }
                        }
                        if (relay_switch_delay) relay_switch_delay--;
                        //
                        i=0;
                        // if the crystal would be exact then we
                        // would be 303.0609 seconds behind every day.
                        // 24*3600 - 24*3600 * 0.9965=303.0609
                        // We compensate 12 sec every hour:
                        // 303.0609 - 12*24=303.0609 - 288=15.069
                        // That is every day we are 15.069 sec behind
                        // We compensate that at 23:30:00 with 15 sec
                        // That means we are 0.069 sec behind per day.
                        if (time[2]>59){ // sec
                                time[2]=0;
                                time[1]++; // count up min
                                gRecMin++;
                                // store graph data 
                                if (gRecMin==((uint16_t)rec_interval*10)){
                                        gRecMin=0; 
                                        store_graph_dat();
                                }
                                // we check if the relay was on
                                // for 5 min and then we record a measurement
                                // for the pool sensor:
                                if (relaystate==1){
                                        // we limit the up counting of relay_on_5_min
                                        // to prevent overflow
                                        if (relay_on_5_min<61){
                                                relay_on_5_min++;
                                        }
                                }else{
                                        relay_on_5_min=0;
                                }
                                i=1; // marker: minutes have changed
                                // count gratutious_arp_cnt down every min:
                                if (gratutious_arp_cnt) gratutious_arp_cnt--;
                        }
                        if (time[1]>59){ // min
                                time[1]=0;
                                time[0]++; // count up hour
                                if (pool_sensor_hours_last_read < 0xffff) pool_sensor_hours_last_read++;
                                // We calibrate the clock every hour a bit:
                                // our clock is a bit behind.
                                time[2]+=12; // adjust/compensate
                        }
                        // once a day at 23:30
                        if (time[0]==23 && time[1] == 30 && time[2]<4){
                                time[2]+=15; // adjust/compensate
                        }
                        if (time[0]>23){ // hours
                                time[0]=0;
                                days++;
                        }
                        // every 10 days:
                        if (days==10){
                                time[2]+=CLOCK_CAL; // adjust/compensate 
                                days=0; // reset to zero otherwise we would do +CLOCK_CAL for a whole day
                        }
                        if (i){ // the values of the minutes did change
                                if (relay_on_5_min>=5){
                                        // relay on for 5min or more (this code will run every minute)
                                        if(pool_sensor_number<gNsensors){
                                                pool_sensor_time[0]=time[0];
                                                pool_sensor_time[1]=time[1];
                                                pool_sensor_time[2]=0; // we don't need sec
                                                pool_sensor_Tempdata=gTempdata[pool_sensor_number];
                                                pool_sensor_hours_last_read=0;
                                        }
                                }
                                relay_change_check();
                        }
                        // Make this work with broken wlan/wifi equipment.
                        // Some of those wifi bridges or wifi client APs
                        // need once in a while a packet otherwise they
                        // forget the MAC address of the connected device and
                        // there is a flaw in those devices such that they
                        // do not broadcast in case of unknown MAC.
                        if (time[2]==15 && gratutious_arp_cnt==0){ // we use time[2], to have an inital arp shortly after power-on
                                gratutious_arp_cnt=5; // every 5min
                                gratutious_arp(buf);
                        }
                        // a button connected between PD6 and GND to
                        // manually change the state of the relay:
                        if (debounce==0 && bit_is_clear(PIND,PIND6)){
                                if (relaystate){
                                        RELAYOFF;
                                        LEDOFF;
                                        relaystate=0;
                                }else{
                                        RELAYON;
                                        LEDON;
                                        relaystate=1;
                                }
                                debounce=0x5fff;
                        }
                        if (debounce) debounce--; // count down button debounce
                        //
                        // we need at least 750ms time between 
                        // start of measurement and reading
                        if (gMeasurementTimer==2 && state==0 ){ 
                                read_temp_meas();
                                state=1;
                        }
                        if (gMeasurementTimer==3 && state==1){ 
                                start_temp_meas();
                                state=2;
                        }
                        if (gMeasurementTimer==5){ // every 5 sec new measurement:
                                gMeasurementTimer=0;
                                state=0;
                        }
                        continue;
                }
                // Cut the size for security reasons. If we are almost at the 
                // end of the buffer then there is a zero but normally there is
                // a lot of room and we can cut down the processing time as 
                // correct URLs should be short in our case. If dat_p is already
                // close to the end then the buffer is terminated already.
                if ((dat_p+65) < BUFFER_SIZE){
                        buf[dat_p+65]='\0';
                }
                // analyse the url and do possible port changes:
                // move one char ahead:
                i=analyse_get_url((char *)&(buf[dat_p+5]));
                // for possible status codes see:
                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                if (i==-1){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1><a href=/>continue</a>\n"));
                        goto SENDTCP;
                }
                if (i==0){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 404 Not Found\r\nContent-Type: text/html\r\n\r\n<h1>404 Not Found</h1>"));
                        goto SENDTCP;
                }
                if (i==10){
                        goto SENDTCP;
                }
                // just display the status:
                // normally i==1
                gPlen=print_webpage_main();
                //
SENDTCP:
                www_server_reply(buf,gPlen); // send web page data
                // tcp port www end
                //
        }
        return (0);
}
